#include <stdlib.h>
#include <stdio.h>
#include "Life.h"
#include "mpi.h"
#include "mpidata.h"
#include<stdlib.h>
#include<limits.h>

#include <X11/Xlib.h> 
#include <assert.h>   
#include<sys/time.h>
#include<sys/resource.h>
#include <unistd.h>   
#define NIL (0)

mpidata g_mpi;

int row_start = -1;
int row_end = -1;

/*Global variables**/

static int bufferLength;

int main(int argc, char ** argv) {

    extern void setup_mpi(int*, char***, mpidata *);
    extern void finalize_mpi();

    // Set up MPI
    setup_mpi(&argc,&argv,&g_mpi);
  
    int rank=g_mpi.rank;
    int size=g_mpi.size;


    bool done=false;
    int count=0;  
    double tempo_gasto;    
    struct rusage usage, prev_usage;

    int north, south;

    // defaults
    int ngrid;
    int ncols;
    int nrows;
    int max_count;
   
    /*Inicializacao dos indices*/
    
    int slice;
    int my_max_iter;
    char *fname;    
    
    char ** grid;
    char ** next_grid;
    
    // command line arguments
    if (rank == 0) {
      if (argc < 5) {
          printf("<Arquivo> <linhas> <colunas> <iteracoes>\n");
          finalize_mpi();
      }
      
      fname = argv[1];
      sscanf(argv[2],"%d",&nrows);
      sscanf(argv[3],"%d",&ncols);
      sscanf(argv[4],"%d",&max_count);
      
      if (nrows%size != 0) {
        printf ("O nmero de linhas deve ser mltiplo de nmero de processadores\n");
        exit(1);
      }
    }

    MPI_Bcast(&nrows,1,MPI_INT,0,MPI_COMM_WORLD);
    MPI_Bcast(&ncols,1,MPI_INT,0,MPI_COMM_WORLD);
    MPI_Bcast(&max_count,1,MPI_INT,0,MPI_COMM_WORLD);
    
    slice = nrows/size;
    bufferLength = ncols+2;

    allocate_grid(ncols,nrows, &grid);
    allocate_grid(ncols,nrows, &next_grid);
    
    if (rank == 0) {

      if (fname == NULL) {
        printf ("arquivo inexistente\n");
        return -1;
      }

      (void)loadData(fname, nrows, ncols, grid);
      
      printGrid(ncols, nrows, grid); /*Imprimindo grade inicial*/
      
      //tinicial = (double) (clock()/CLOCKS_PER_SEC); //INICIO DO TEMPO
      getrusage(RUSAGE_SELF, &prev_usage);
      
      for (int i=0; i<size; i++) {
	
        my_max_iter = slice*(i+1);
	
        for (int j=i*slice+1; j<=my_max_iter; j++)
	  MPI_Send(grid[j], bufferLength, MPI_CHAR, i, TAG, MPI_COMM_WORLD);
      }
    }
    
    north=row_start = slice*rank+1;
    south=row_end = slice*(rank+1);
    
    for (int i = row_start; i<=row_end; i++) {
      MPI_Recv (grid[i], bufferLength, MPI_CHAR, 0, TAG,MPI_COMM_WORLD, &g_mpi.status);
    }
    
    while(count < max_count) {
	
        do_step(rank, size, ncols, grid, next_grid);
	
        for (int i=0;i<nrows+2;i++) {
	  for (int j=0;j<ncols+2;j++) {
	    grid[i][j]=next_grid[i][j];
	  }
        }
	count++;
    }

    for (int i = north; i<=south; i++)
      MPI_Send (grid[i],bufferLength,MPI_CHAR,0,TAG,MPI_COMM_WORLD);
    
    if (rank == 0) {
      
      for (int i=0; i<size; i++) {
	
        my_max_iter = slice*(i+1);
       
        for (int j=i*slice+1; j<=my_max_iter; j++) {
	  MPI_Recv(grid[j], bufferLength, MPI_CHAR, i, TAG,MPI_COMM_WORLD, &g_mpi.status); 
	}
      }
      
      printGrid(ncols,nrows,grid); /*Imprimindo grade final*/
     
      //tfinal =(double) (clock()/CLOCKS_PER_SEC); //FIM DO TEMPO

      getrusage(RUSAGE_SELF, &usage);
      tempo_gasto = (((double)usage.ru_utime.tv_sec+((double)usage.ru_utime.tv_usec/1e06))-((double)prev_usage.ru_utime.tv_sec+((double)prev_usage.ru_utime.tv_usec/10e06)));
  printf ("O tempo de processamento de usuario  de : %f\n", tempo_gasto);

    }

    cleanup_grid(ncols,nrows, &grid);
    cleanup_grid(ncols,nrows, &next_grid);
    
    printf ("rank %d\n", rank);

    finalize_mpi();

    return 0;
}

/* param :
 *      row_start - primeira linha com conteudo relevante
 *      row_end - ultima linha com conteudo relevante
 *      rank - identificador de processador
 *      size - numero de processadores
 *      grid - grade com configuracao atual para o processador i
 *      next_grid - grade que conterah nova conf. apos aplicacao de regras
 *     
 * Efeito :
 *      Aplica um passo de jogo(regras) a cada processador levando em conta a 
 *     conversa entre este e os vizinhos.
 */

void do_step(int rank, int size, int ncols, char ** grid, char ** next_grid) {
  
  int left_rank = (rank-1+size)%size;
  int right_rank = (rank+1)%size;
  
  if (rank == 0) {
    MPI_Send(grid[row_end],bufferLength,MPI_CHAR,right_rank,MPIDATA_TOLEFT, 
    MPI_COMM_WORLD);
    row_end++;
    MPI_Recv(grid[row_end],bufferLength,MPI_CHAR,right_rank,
	     MPIDATA_TOLEFT,MPI_COMM_WORLD, &g_mpi.status);
    
  }

  else if (rank == size-1) {
    MPI_Send(grid[row_start],bufferLength,MPI_CHAR,left_rank,MPIDATA_TOLEFT, 
	     MPI_COMM_WORLD);
    row_start--;
    MPI_Recv(grid[row_start],bufferLength,MPI_CHAR,left_rank,
	     MPIDATA_TOLEFT,MPI_COMM_WORLD, &g_mpi.status);
    
  }
  else {
    MPI_Send(grid[row_start],bufferLength,MPI_CHAR,left_rank,MPIDATA_TOLEFT, 
	     MPI_COMM_WORLD);
    row_start--;
    MPI_Recv(grid[row_start],bufferLength,MPI_CHAR,left_rank,
	     MPIDATA_TOLEFT,MPI_COMM_WORLD, &g_mpi.status);

    MPI_Send(grid[row_end],bufferLength,MPI_CHAR,right_rank,MPIDATA_TOLEFT, 
	     MPI_COMM_WORLD);
    row_end++;
    MPI_Recv(grid[row_end],bufferLength,MPI_CHAR,right_rank,
	     MPIDATA_TOLEFT,MPI_COMM_WORLD, &g_mpi.status);
     
  }  

  /***************************************************/
  //update
  
  for (int i=row_start;i<=row_end;i++) {
    for (int j=1;j<=ncols;j++) {
      int neighbors=0;
      for (int k=i-1;k<=i+1; k++) {
	for (int l=j-1;l<=j+1; l++) {
	  if (!(k==i&&l==j)&&(grid[k][l]== '1')) {
	      neighbors++;
	  }
	}
	//if (neighbors>3) continue;
      }//fim do calculo das vizinhas


      //Se a celula eh viva
      if (grid[i][j] == '1') {
        if ((neighbors != 2) && (neighbors != 3)) next_grid[i][j] = '0';
        else next_grid[i][j] = '1';
      }

      //Se a celula eh morta
      else {
        if (neighbors == 3) next_grid[i][j] = '1';
        else next_grid[i][j] = '0';
      }
    }
  }
  if (rank == 0) {row_end--;}
  else if (rank == size-1) {row_start++;}
  else {row_start++; row_end--;}
}

void allocate_grid(int ncols, int nrows, char *** grid){
  (*grid) = new char *[nrows+2];
  for (int i=0; i<nrows+2;i++) {
    (*grid)[i] = new char[ncols+2];
    for (int j=0;j<ncols+2;j++) {
      (*grid)[i][j]='0';
    }
  }
}

void cleanup_grid(int ncols, int nrows, char *** grid){
    for (int i=0;i<nrows+2;i++) {
        delete (*grid)[i];
    }
    delete (*grid);
}

/* Leitura de arquivos de dados*/
int loadData(char *fname, int nrows, int ncols, char **grid)
{
  FILE *fp; 

  fp = fopen (fname, "r");
  if (fp == NULL) {
    printf ("Erro na abertura do arquivo %s.\n", fname);
    return EXIT_FAILURE;
  }

  printf("\nLoading data from file '%s'\n", fname);
  
  /* leitura da matriz */
  for (int i = 1; i < nrows+1; i++) {
    for (int j = 1; j < ncols+1; j++)
      if (fscanf (fp, "%c ", &grid[i][j]) != 1) {
	printf ("%s  no est completo ou "
		"possui algum smbolo invlido.\n", fname);
        return -1;
      }
  }
  fclose (fp);
  return 0;
}


void printGrid(int cols, int rows, char **grid) {
  for (int i=0; i<rows+2; i++) {
    for (int j=0; j<cols+2; j++) {
      if (i == 0 || j == 0 || (i == rows+1) || (j == cols+1))
	printf ("%c", grid[i][j]);
      else if (grid[i][j] == '1') printf ("V");
      else printf ("M");
    }
    printf("\n");
  }
  printf("\n\n");
}

// X information, at some point this should be cleaned up so
// that it does not use global variables

// setupWindow modified from the tutorial on
// http://tronche.com/gui/x/xlib-tutorial/
// by Christophe Tronche


Display *dpy;
int blackColor;
int whiteColor;
Window w;
GC gc;
Pixmap buffer;
Colormap theColormap;
int numXGrayscale=10;
XColor Xgrayscale[10];
int IMAGE_WIDTH=LIFE_IMAGE_WIDTH;
int IMAGE_HEIGHT=LIFE_IMAGE_HEIGHT;

void setupWindow(int ncols, int nrows) {
      // Open the display

      dpy = XOpenDisplay(NIL);
      assert(dpy);

      // Get some colors

      blackColor = BlackPixel(dpy, DefaultScreen(dpy));
      whiteColor = WhitePixel(dpy, DefaultScreen(dpy));

      // Create the window
      if (nrows>ncols) {
	printf ("Atencao vc passou mais linhas que colunas\n");
	IMAGE_WIDTH = (int)((double)LIFE_IMAGE_WIDTH*(double)ncols/(double)nrows);
	IMAGE_HEIGHT = LIFE_IMAGE_HEIGHT;
      } else {
	IMAGE_HEIGHT = (int)((double)LIFE_IMAGE_HEIGHT*(double)nrows/(double)ncols);
	IMAGE_WIDTH = LIFE_IMAGE_WIDTH;
      }
      
      
      w = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 0, 0, 
			      IMAGE_WIDTH, IMAGE_HEIGHT, 0, blackColor,
			      blackColor);
      buffer = XCreatePixmap(dpy,DefaultRootWindow(dpy),
			     IMAGE_WIDTH,IMAGE_HEIGHT,DefaultDepth(dpy,
								   DefaultScreen(dpy)));
      
      theColormap = XCreateColormap(dpy, DefaultRootWindow(dpy),
				    DefaultVisual(dpy,DefaultScreen(dpy)), AllocNone);
      
      for (int i=0;i<numXGrayscale;i++) {
	int color = (int)((double)i*35535.0/(double)numXGrayscale)+30000;
	Xgrayscale[i].red=color;
	Xgrayscale[i].green=color;
	Xgrayscale[i].blue=color;
	XAllocColor(dpy,theColormap,&(Xgrayscale[i]));
      }
      
      // We want to get MapNotify events
      
      XSelectInput(dpy, w, StructureNotifyMask);
      
      // "Map" the window (that is, make it appear on the screen)
      
      XMapWindow(dpy, w);
      
      // Create a "Graphics Context"
      
      gc = XCreateGC(dpy, w, 0, NIL);
      
      // Tell the GC we draw using the white color
      
      XSetForeground(dpy, gc, whiteColor);
      
      // Wait for the MapNotify event
      
      for(;;) {
	XEvent e;
	XNextEvent(dpy, &e);
	if (e.type == MapNotify)
	  break;
      }
      
}

void doDraw(int rank, int ncols, int nrows, int ** grid) {

    int x1,x2,y1,y2; 
    char string[2];
    sprintf(string,"%d",rank);
    
    XSetForeground(dpy, gc, blackColor);
    XFillRectangle(dpy,buffer,gc,0,0,IMAGE_WIDTH,IMAGE_HEIGHT);
    int rect_width=(int)((double)IMAGE_WIDTH/(double)(ncols+1));
    int rect_height=(int)((double)IMAGE_HEIGHT/(double)(nrows+1));
    for (int i=1;i<nrows+1;i++) {
        x1 = (int)((double)(i-1)/(double)(ncols+1)*(double)IMAGE_WIDTH);
        for (int j=1;j<ncols+1;j++) {
            y1 = (int)((double)(j-1)/(double)(nrows+1)*
                (double)IMAGE_HEIGHT);
            if (grid[i][j]>0) {
                int life =grid[i][j];
                if (life>numXGrayscale-1) life=numXGrayscale-1;
                XSetForeground(dpy, gc, Xgrayscale[life].pixel);
            } else {
                XSetForeground(dpy, gc, blackColor);
            }
            XFillRectangle(dpy,buffer,gc,x1,y1,rect_width,rect_height);
         }
     }
     XSetForeground(dpy,gc,blackColor);
     XFillRectangle(dpy,buffer,gc,10,10,15,15);
     XSetForeground(dpy,gc,whiteColor);
     XDrawRectangle(dpy,buffer,gc,10,10,15,15);
     XDrawString(dpy,buffer,gc,12,23,string,2);
     
     XCopyArea(dpy, buffer, w, gc, 0, 0,
         IMAGE_WIDTH, IMAGE_HEIGHT,  0, 0);
     XFlush(dpy);
}


void setup_mpi(int *argc,char*** argv, mpidata* my_mpi) {
    MPI_Init(argc,argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &my_mpi->rank);
    MPI_Comm_size(MPI_COMM_WORLD, &my_mpi->size);
}

void finalize_mpi() {
    MPI_Finalize();
}

